Please fix the code based on the following feedback

# Review comments
{{ review_comments }}

# Review threads
{{ review_threads }}

# Review thread files
{{ files }}

# PR Thread Comments
{{ thread_context }}